@@ -7,7 +7,8 @@
 
 #include <cstdint>
 
-#include "src/base/functional.h"
+#include "include/v8-internal.h"
+#include "src/base/hashing.h"
 #include "src/base/logging.h"
 #include "src/codegen/machine-type.h"
 #include "src/compiler/turboshaft/utils.h"
@@ -28,6 +29,7 @@ class MaybeRegisterRepresentation {
     kTagged,
     kCompressed,
     kSimd128,
+    kSimd256,
     kNone,  // No register representation.
   };
 
@@ -51,6 +53,15 @@ class MaybeRegisterRepresentation {
     return MaybeRegisterRepresentation(Enum::kWord64);
   }
 
+  static constexpr MaybeRegisterRepresentation WordPtr() {
+    if constexpr (kSystemPointerSize == 4) {
+      return Word32();
+    } else {
+      DCHECK_EQ(kSystemPointerSize, 8);
+      return Word64();
+    }
+  }
+
   static constexpr MaybeRegisterRepresentation Float32() {
     return MaybeRegisterRepresentation(Enum::kFloat32);
   }
@@ -71,17 +82,12 @@ class MaybeRegisterRepresentation {
     return MaybeRegisterRepresentation(Enum::kSimd128);
   }
 
-  static constexpr MaybeRegisterRepresentation None() {
-    return MaybeRegisterRepresentation(Enum::kNone);
+  static constexpr MaybeRegisterRepresentation Simd256() {
+    return MaybeRegisterRepresentation(Enum::kSimd256);
   }
 
-  static constexpr MaybeRegisterRepresentation PointerSized() {
-    if constexpr (kSystemPointerSize == 4) {
-      return Word32();
-    } else {
-      DCHECK_EQ(kSystemPointerSize, 8);
-      return Word64();
-    }
+  static constexpr MaybeRegisterRepresentation None() {
+    return MaybeRegisterRepresentation(Enum::kNone);
   }
 
   constexpr bool IsWord() const {
@@ -94,6 +100,7 @@ class MaybeRegisterRepresentation {
       case Enum::kTagged:
       case Enum::kCompressed:
       case Enum::kSimd128:
+      case Enum::kSimd256:
       case Enum::kNone:
         return false;
     }
@@ -109,6 +116,23 @@ class MaybeRegisterRepresentation {
       case Enum::kTagged:
       case Enum::kCompressed:
       case Enum::kSimd128:
+      case Enum::kSimd256:
+      case Enum::kNone:
+        return false;
+    }
+  }
+
+  constexpr bool IsTaggedOrCompressed() const {
+    switch (*this) {
+      case Enum::kTagged:
+      case Enum::kCompressed:
+        return true;
+      case Enum::kWord32:
+      case Enum::kWord64:
+      case Enum::kFloat32:
+      case Enum::kFloat64:
+      case Enum::kSimd128:
+      case Enum::kSimd256:
       case Enum::kNone:
         return false;
     }
@@ -125,6 +149,7 @@ class MaybeRegisterRepresentation {
       case Enum::kTagged:
       case Enum::kCompressed:
       case Enum::kSimd128:
+      case Enum::kSimd256:
       case Enum::kNone:
         UNREACHABLE();
     }
@@ -146,6 +171,8 @@ class MaybeRegisterRepresentation {
         return MachineRepresentation::kCompressed;
       case Simd128():
         return MachineRepresentation::kSimd128;
+      case Simd256():
+        return MachineRepresentation::kSimd256;
       case None():
         UNREACHABLE();
     }
@@ -167,6 +194,8 @@ class MaybeRegisterRepresentation {
         return kSystemPointerSize;
       case Simd128():
         return 128;
+      case Simd256():
+        return 256;
       case None():
         UNREACHABLE();
     }
@@ -189,6 +218,7 @@ class RegisterRepresentation : public MaybeRegisterRepresentation {
     kCompressed =
         static_cast<int>(MaybeRegisterRepresentation::Enum::kCompressed),
     kSimd128 = static_cast<int>(MaybeRegisterRepresentation::Enum::kSimd128),
+    kSimd256 = static_cast<int>(MaybeRegisterRepresentation::Enum::kSimd256),
   };
 
   explicit constexpr RegisterRepresentation(Enum value)
@@ -211,6 +241,11 @@ class RegisterRepresentation : public MaybeRegisterRepresentation {
   static constexpr RegisterRepresentation Word64() {
     return RegisterRepresentation(Enum::kWord64);
   }
+  // The equivalent of intptr_t/uintptr_t: An integral type with the same size
+  // as machine pointers.
+  static constexpr RegisterRepresentation WordPtr() {
+    return RegisterRepresentation(MaybeRegisterRepresentation::WordPtr());
+  }
   static constexpr RegisterRepresentation Float32() {
     return RegisterRepresentation(Enum::kFloat32);
   }
@@ -227,16 +262,14 @@ class RegisterRepresentation : public MaybeRegisterRepresentation {
   static constexpr RegisterRepresentation Compressed() {
     return RegisterRepresentation(Enum::kCompressed);
   }
-  // The equivalent of intptr_t/uintptr_t: An integral type with the same size
-  // as machine pointers.
-  static constexpr RegisterRepresentation PointerSized() {
-    return RegisterRepresentation(MaybeRegisterRepresentation::PointerSized());
-  }
   static constexpr RegisterRepresentation Simd128() {
     return RegisterRepresentation(Enum::kSimd128);
   }
+  static constexpr RegisterRepresentation Simd256() {
+    return RegisterRepresentation(Enum::kSimd256);
+  }
 
-  static RegisterRepresentation FromMachineRepresentation(
+  static constexpr RegisterRepresentation FromMachineRepresentation(
       MachineRepresentation rep) {
     switch (rep) {
       case MachineRepresentation::kBit:
@@ -249,27 +282,68 @@ class RegisterRepresentation : public MaybeRegisterRepresentation {
       case MachineRepresentation::kTaggedSigned:
       case MachineRepresentation::kTaggedPointer:
       case MachineRepresentation::kTagged:
+      case MachineRepresentation::kProtectedPointer:
         return Tagged();
       case MachineRepresentation::kCompressedPointer:
       case MachineRepresentation::kCompressed:
         return Compressed();
+      case MachineRepresentation::kFloat16:
       case MachineRepresentation::kFloat32:
         return Float32();
       case MachineRepresentation::kFloat64:
         return Float64();
       case MachineRepresentation::kSimd128:
         return Simd128();
+      case MachineRepresentation::kSimd256:
+        return Simd256();
       case MachineRepresentation::kMapWord:
+        // Turboshaft does not support map packing.
+        DCHECK(!V8_MAP_PACKING_BOOL);
+        return RegisterRepresentation::Tagged();
       case MachineRepresentation::kIndirectPointer:
       case MachineRepresentation::kSandboxedPointer:
+        // TODO(saelo/jkummerow): This is suspicious: after resolving the
+        // indirection, we have a Tagged pointer.
+        return WordPtr();
       case MachineRepresentation::kNone:
-      case MachineRepresentation::kSimd256:
+      case MachineRepresentation::kFloat16RawBits:
         UNREACHABLE();
     }
   }
 
+  static constexpr RegisterRepresentation FromMachineType(MachineType type) {
+    return FromMachineRepresentation(type.representation());
+  }
+
+  static constexpr RegisterRepresentation FromCTypeInfo(
+      CTypeInfo t, CFunctionInfo::Int64Representation int64_repr) {
+    if (t.GetType() == CTypeInfo::Type::kVoid ||
+        t.GetType() == CTypeInfo::Type::kPointer) {
+      return RegisterRepresentation::Tagged();
+    } else if (t.GetType() == CTypeInfo::Type::kInt64 ||
+               t.GetType() == CTypeInfo::Type::kUint64) {
+      if (int64_repr == CFunctionInfo::Int64Representation::kBigInt) {
+        return RegisterRepresentation::Word64();
+      } else {
+        DCHECK_EQ(int64_repr, CFunctionInfo::Int64Representation::kNumber);
+        return RegisterRepresentation::Float64();
+      }
+    } else {
+      return RegisterRepresentation::FromMachineType(
+          MachineType::TypeForCType(t));
+    }
+  }
+
   constexpr bool AllowImplicitRepresentationChangeTo(
-      RegisterRepresentation dst_rep) const;
+      RegisterRepresentation dst_rep, bool graph_created_from_turbofan) const;
+
+  constexpr RegisterRepresentation MapTaggedToWord() const {
+    if (this->value() == RegisterRepresentation::Tagged()) {
+      return COMPRESS_POINTERS_BOOL ? RegisterRepresentation::Word32()
+                                    : RegisterRepresentation::WordPtr();
+    }
+    return *this;
+  }
 };
 
 V8_INLINE constexpr bool operator==(MaybeRegisterRepresentation a,
@@ -286,18 +360,12 @@ V8_INLINE size_t hash_value(MaybeRegisterRepresentation rep) {
 }
 
 constexpr bool RegisterRepresentation::AllowImplicitRepresentationChangeTo(
-    RegisterRepresentation dst_rep) const {
+    RegisterRepresentation dst_rep, bool graph_created_from_turbofan) const {
   if (*this == dst_rep) {
     return true;
   }
   switch (dst_rep.value()) {
     case RegisterRepresentation::Word32():
-      // TODO(mliedtke): Remove this once JS graph building and JS reducers
-      // always produce explicit truncations.
-      // We allow implicit 64- to 32-bit truncation.
-      if (*this == RegisterRepresentation::Word64()) {
-        return true;
-      }
       // We allow implicit tagged -> untagged conversions.
       // Even without pointer compression, we use `Word32And` for Smi-checks on
       // tagged values.
@@ -305,6 +373,14 @@ constexpr bool RegisterRepresentation::AllowImplicitRepresentationChangeTo(
                           RegisterRepresentation::Compressed())) {
         return true;
       }
+      if (graph_created_from_turbofan &&
+          *this == RegisterRepresentation::Word64()) {
+        // TODO(12783): Remove this once Turboshaft graphs are not constructed
+        // via Turbofan any more. Unfortunately Turbofan has many implicit
+        // truncations which are hard to fix. Still, for wasm it is required
+        // that truncations in Turboshaft are explicit.
+        return true;
+      }
       break;
     case RegisterRepresentation::Word64():
       // We allow implicit tagged -> untagged conversions.
@@ -316,14 +392,14 @@ constexpr bool RegisterRepresentation::AllowImplicitRepresentationChangeTo(
     case RegisterRepresentation::Tagged():
       // We allow implicit untagged -> tagged conversions. This is only safe for
       // Smi values.
-      if (*this == RegisterRepresentation::PointerSized()) {
+      if (*this == RegisterRepresentation::WordPtr()) {
         return true;
       }
       break;
     case RegisterRepresentation::Compressed():
       // Compression is a no-op.
       if (*this == any_of(RegisterRepresentation::Tagged(),
-                          RegisterRepresentation::PointerSized(),
+                          RegisterRepresentation::WordPtr(),
                           RegisterRepresentation::Word32())) {
         return true;
       }
@@ -334,7 +410,8 @@ constexpr bool RegisterRepresentation::AllowImplicitRepresentationChangeTo(
   return false;
 }
 
-std::ostream& operator<<(std::ostream& os, MaybeRegisterRepresentation rep);
+V8_EXPORT_PRIVATE std::ostream& operator<<(std::ostream& os,
+                                           MaybeRegisterRepresentation rep);
 
 template <typename T>
 struct MultiSwitch<
@@ -369,8 +446,8 @@ class WordRepresentation : public RegisterRepresentation {
     return WordRepresentation(Enum::kWord64);
   }
 
-  static constexpr WordRepresentation PointerSized() {
-    return WordRepresentation(RegisterRepresentation::PointerSized());
+  static constexpr WordRepresentation WordPtr() {
+    return WordRepresentation(RegisterRepresentation::WordPtr());
   }
 
   constexpr Enum value() const {
@@ -421,6 +498,10 @@ class FloatRepresentation : public RegisterRepresentation {
   explicit constexpr FloatRepresentation(Enum value)
       : RegisterRepresentation(
             static_cast<RegisterRepresentation::Enum>(value)) {}
+  explicit constexpr FloatRepresentation(RegisterRepresentation rep)
+      : FloatRepresentation(static_cast<Enum>(rep.value())) {
+    DCHECK(rep.IsFloat());
+  }
   FloatRepresentation() = default;
 
   constexpr Enum value() const {
@@ -440,14 +521,20 @@ class MemoryRepresentation {
     kUint32,
     kInt64,
     kUint64,
+    kFloat16,
     kFloat32,
     kFloat64,
     kAnyTagged,
     kTaggedPointer,
     kTaggedSigned,
+    kAnyUncompressedTagged,
+    kUncompressedTaggedPointer,
+    kUncompressedTaggedSigned,
+    kProtectedPointer,
     kIndirectPointer,
     kSandboxedPointer,
     kSimd128,
+    kSimd256
   };
 
   explicit constexpr MemoryRepresentation(Enum value) : value_(value) {}
@@ -485,6 +572,16 @@ class MemoryRepresentation {
   static constexpr MemoryRepresentation Uint64() {
     return MemoryRepresentation(Enum::kUint64);
   }
+  static constexpr MemoryRepresentation UintPtr() {
+    if constexpr (Is64()) {
+      return Uint64();
+    } else {
+      return Uint32();
+    }
+  }
+  static constexpr MemoryRepresentation Float16() {
+    return MemoryRepresentation(Enum::kFloat16);
+  }
   static constexpr MemoryRepresentation Float32() {
     return MemoryRepresentation(Enum::kFloat32);
   }
@@ -500,45 +597,29 @@ class MemoryRepresentation {
   static constexpr MemoryRepresentation TaggedSigned() {
     return MemoryRepresentation(Enum::kTaggedSigned);
   }
+  static constexpr MemoryRepresentation AnyUncompressedTagged() {
+    return MemoryRepresentation(Enum::kAnyUncompressedTagged);
+  }
+  static constexpr MemoryRepresentation UncompressedTaggedPointer() {
+    return MemoryRepresentation(Enum::kUncompressedTaggedPointer);
+  }
+  static constexpr MemoryRepresentation UncompressedTaggedSigned() {
+    return MemoryRepresentation(Enum::kUncompressedTaggedSigned);
+  }
+  static constexpr MemoryRepresentation ProtectedPointer() {
+    return MemoryRepresentation(Enum::kProtectedPointer);
+  }
   static constexpr MemoryRepresentation IndirectPointer() {
     return MemoryRepresentation(Enum::kIndirectPointer);
   }
   static constexpr MemoryRepresentation SandboxedPointer() {
     return MemoryRepresentation(Enum::kSandboxedPointer);
   }
-  static constexpr MemoryRepresentation PointerSized() {
-    if constexpr (kSystemPointerSize == 4) {
-      return Uint32();
-    } else {
-      DCHECK_EQ(kSystemPointerSize, 8);
-      return Uint64();
-    }
-  }
   static constexpr MemoryRepresentation Simd128() {
     return MemoryRepresentation(Enum::kSimd128);
   }
-
-  bool IsWord() const {
-    switch (*this) {
-      case Int8():
-      case Uint8():
-      case Int16():
-      case Uint16():
-      case Int32():
-      case Uint32():
-      case Int64():
-      case Uint64():
-        return true;
-      case Float32():
-      case Float64():
-      case AnyTagged():
-      case TaggedPointer():
-      case TaggedSigned():
-      case IndirectPointer():
-      case SandboxedPointer():
-      case Simd128():
-        return false;
-    }
+  static constexpr MemoryRepresentation Simd256() {
+    return MemoryRepresentation(Enum::kSimd256);
   }
 
   bool IsSigned() const {
@@ -553,19 +634,29 @@ class MemoryRepresentation {
       case Uint32():
       case Uint64():
         return false;
+      case Float16():
       case Float32():
       case Float64():
       case AnyTagged():
       case TaggedPointer():
       case TaggedSigned():
+      case AnyUncompressedTagged():
+      case UncompressedTaggedPointer():
+      case UncompressedTaggedSigned():
+      case ProtectedPointer():
       case IndirectPointer():
       case SandboxedPointer():
       case Simd128():
+      case Simd256():
         UNREACHABLE();
     }
   }
 
-  bool IsTagged() const {
+  // This predicate is used in particular to decide which load/store ops
+  // have to deal with pointer compression. Indirect/sandboxed pointers,
+  // while they resolve to tagged pointers, return {false} because they
+  // use incompatible compression schemes.
+  bool IsCompressibleTagged() const {
     switch (*this) {
       case AnyTagged():
       case TaggedPointer():
@@ -579,34 +670,17 @@ class MemoryRepresentation {
       case Uint16():
       case Uint32():
       case Uint64():
+      case Float16():
       case Float32():
       case Float64():
+      case AnyUncompressedTagged():
+      case UncompressedTaggedPointer():
+      case UncompressedTaggedSigned():
       case IndirectPointer():
+      case ProtectedPointer():
       case SandboxedPointer():
       case Simd128():
-        return false;
-    }
-  }
-
-  bool CanBeTaggedPointer() const {
-    switch (*this) {
-      case AnyTagged():
-      case TaggedPointer():
-        return true;
-      case TaggedSigned():
-      case Int8():
-      case Int16():
-      case Int32():
-      case Int64():
-      case Uint8():
-      case Uint16():
-      case Uint32():
-      case Uint64():
-      case Float32():
-      case Float64():
-      case IndirectPointer():
-      case SandboxedPointer():
-      case Simd128():
+      case Simd256():
         return false;
     }
   }
@@ -623,6 +697,7 @@ class MemoryRepresentation {
       case Int64():
       case Uint64():
         return RegisterRepresentation::Word64();
+      case Float16():
       case Float32():
         return RegisterRepresentation::Float32();
       case Float64():
@@ -630,13 +705,18 @@ class MemoryRepresentation {
       case AnyTagged():
       case TaggedPointer():
       case TaggedSigned():
-        return RegisterRepresentation::Tagged();
+      case AnyUncompressedTagged():
+      case UncompressedTaggedPointer():
+      case UncompressedTaggedSigned():
       case IndirectPointer():
+      case ProtectedPointer():
         return RegisterRepresentation::Tagged();
       case SandboxedPointer():
         return RegisterRepresentation::Word64();
       case Simd128():
         return RegisterRepresentation::Simd128();
+      case Simd256():
+        return RegisterRepresentation::Simd256();
     }
   }
 
@@ -655,6 +735,8 @@ class MemoryRepresentation {
         return AnyTagged();
       case RegisterRepresentation::Simd128():
         return Simd128();
+      case RegisterRepresentation::Simd256():
+        return Simd256();
       case RegisterRepresentation::Compressed():
         UNREACHABLE();
     }
@@ -691,6 +773,8 @@ class MemoryRepresentation {
         return MachineType::Int64();
       case Uint64():
         return MachineType::Uint64();
+      case Float16():
+        return MachineType::Float16();
       case Float32():
         return MachineType::Float32();
       case Float64():
@@ -701,12 +785,22 @@ class MemoryRepresentation {
         return MachineType::TaggedPointer();
       case TaggedSigned():
         return MachineType::TaggedSigned();
+      case AnyUncompressedTagged():
+        return MachineType::AnyTagged();
+      case UncompressedTaggedPointer():
+        return MachineType::TaggedPointer();
+      case UncompressedTaggedSigned():
+        return MachineType::TaggedSigned();
+      case ProtectedPointer():
+        return MachineType::ProtectedPointer();
       case IndirectPointer():
         return MachineType::IndirectPointer();
       case SandboxedPointer():
         return MachineType::SandboxedPointer();
       case Simd128():
         return MachineType::Simd128();
+      case Simd256():
+        return MachineType::Simd256();
     }
   }
 
@@ -728,10 +822,14 @@ class MemoryRepresentation {
         // Turboshaft does not support map packing.
         DCHECK(!V8_MAP_PACKING_BOOL);
         return TaggedPointer();
+      case MachineRepresentation::kProtectedPointer:
+        return ProtectedPointer();
       case MachineRepresentation::kIndirectPointer:
         return IndirectPointer();
       case MachineRepresentation::kTagged:
         return AnyTagged();
+      case MachineRepresentation::kFloat16:
+        return Float16();
       case MachineRepresentation::kFloat32:
         return Float32();
       case MachineRepresentation::kFloat64:
@@ -740,11 +838,13 @@ class MemoryRepresentation {
         return SandboxedPointer();
       case MachineRepresentation::kSimd128:
         return Simd128();
+      case MachineRepresentation::kSimd256:
+        return Simd256();
       case MachineRepresentation::kNone:
       case MachineRepresentation::kBit:
-      case MachineRepresentation::kSimd256:
       case MachineRepresentation::kCompressedPointer:
       case MachineRepresentation::kCompressed:
+      case MachineRepresentation::kFloat16RawBits:
         UNREACHABLE();
     }
   }
@@ -766,6 +866,8 @@ class MemoryRepresentation {
         return TaggedPointer();
       case MachineRepresentation::kTagged:
         return AnyTagged();
+      case MachineRepresentation::kFloat16:
+        return Float16();
       case MachineRepresentation::kFloat32:
         return Float32();
       case MachineRepresentation::kFloat64:
@@ -774,13 +876,16 @@ class MemoryRepresentation {
         return SandboxedPointer();
       case MachineRepresentation::kSimd128:
         return Simd128();
+      case MachineRepresentation::kSimd256:
+        return Simd256();
       case MachineRepresentation::kNone:
       case MachineRepresentation::kMapWord:
       case MachineRepresentation::kBit:
-      case MachineRepresentation::kSimd256:
       case MachineRepresentation::kCompressedPointer:
       case MachineRepresentation::kCompressed:
+      case MachineRepresentation::kProtectedPointer:
       case MachineRepresentation::kIndirectPointer:
+      case MachineRepresentation::kFloat16RawBits:
         UNREACHABLE();
     }
   }
@@ -796,6 +901,7 @@ class MemoryRepresentation {
         return 0;
       case Int16():
       case Uint16():
+      case Float16():
         return 1;
       case Int32():
       case Uint32():
@@ -810,9 +916,16 @@ class MemoryRepresentation {
       case AnyTagged():
       case TaggedPointer():
       case TaggedSigned():
+      case ProtectedPointer():
         return kTaggedSizeLog2;
+      case AnyUncompressedTagged():
+      case UncompressedTaggedPointer():
+      case UncompressedTaggedSigned():
+        return kSystemPointerSizeLog2;
       case Simd128():
         return 4;
+      case Simd256():
+        return 5;
     }
   }
 
@@ -822,10 +935,12 @@ class MemoryRepresentation {
   static constexpr Enum kInvalid = static_cast<Enum>(-1);
 };
 
-V8_INLINE bool operator==(MemoryRepresentation a, MemoryRepresentation b) {
+V8_INLINE constexpr bool operator==(MemoryRepresentation a,
+                                    MemoryRepresentation b) {
   return a.value() == b.value();
 }
-V8_INLINE bool operator!=(MemoryRepresentation a, MemoryRepresentation b) {
+V8_INLINE constexpr bool operator!=(MemoryRepresentation a,
+                                    MemoryRepresentation b) {
   return a.value() != b.value();
 }
 
@@ -833,7 +948,8 @@ V8_INLINE size_t hash_value(MemoryRepresentation rep) {
   return static_cast<size_t>(rep.value());
 }
 
-std::ostream& operator<<(std::ostream& os, MemoryRepresentation rep);
+V8_EXPORT_PRIVATE std::ostream& operator<<(std::ostream& os,
+                                           MemoryRepresentation rep);
 
 }  // namespace v8::internal::compiler::turboshaft
 
