@@ -18,7 +18,6 @@ const {
   kIsExecuting,
   kRequiredModuleSymbol,
 } = require('internal/modules/cjs/loader');
-
 const { imported_cjs_symbol } = internalBinding('symbols');
 
 const assert = require('internal/assert');
@@ -38,7 +37,16 @@ const {
   forceDefaultLoader,
 } = require('internal/modules/esm/utils');
 const { kImplicitTypeAttribute } = require('internal/modules/esm/assert');
-const { ModuleWrap, kEvaluating, kEvaluated } = internalBinding('module_wrap');
+const {
+  ModuleWrap,
+  kEvaluated,
+  kEvaluating,
+  kEvaluationPhase,
+  kInstantiated,
+  kErrored,
+  kSourcePhase,
+  throwIfPromiseRejected,
+} = internalBinding('module_wrap');
 const {
   urlToFilename,
 } = require('internal/modules/helpers');
@@ -53,7 +61,13 @@ let defaultResolve, defaultLoad, defaultLoadSync, importMetaInitializer;
 const { tracingChannel } = require('diagnostics_channel');
 const onImport = tracingChannel('module.import');
 
+let debug = require('internal/util/debuglog').debuglog('esm', (fn) => {
+  debug = fn;
+});
+
 /**
+ * @typedef {import('./hooks.js').HooksProxy} HooksProxy
+ * @typedef {import('./module_job.js').ModuleJobBase} ModuleJobBase
  * @typedef {import('url').URL} URL
  */
 
@@ -84,6 +98,23 @@ function getTranslators() {
   return translators;
 }
 
+/**
+ * Generate message about potential race condition caused by requiring a cached module that has started
+ * async linking.
+ * @param {string} filename Filename of the module being required.
+ * @param {string|undefined} parentFilename Filename of the module calling require().
+ * @returns {string} Error message.
+ */
+function getRaceMessage(filename, parentFilename) {
+  let raceMessage = `Cannot require() ES Module ${filename} because it is not yet fully loaded. `;
+  raceMessage += 'This may be caused by a race condition if the module is simultaneously dynamically ';
+  raceMessage += 'import()-ed via Promise.all(). Try await-ing the import() sequentially in a loop instead.';
+  if (parentFilename) {
+    raceMessage += ` (from ${parentFilename})`;
+  }
+  return raceMessage;
+}
+
 /**
  * @type {HooksProxy}
  * Multiple loader instances exist for various, specific reasons (see code comments at site).
@@ -149,6 +180,7 @@ class ModuleLoader {
    *  to this property and failure to do so will cause undefined
    * behavior when invoking `import.meta.resolve`.
    * @see {ModuleLoader.setCustomizations}
+   * @type {CustomizedModuleLoader}
    */
   #customizations;
 
@@ -202,7 +234,7 @@ class ModuleLoader {
    *
    * Calling this function alters how modules are loaded and should be
    * invoked with care.
-   * @param {object} customizations
+   * @param {CustomizedModuleLoader} customizations
    */
   setCustomizations(customizations) {
     this.#customizations = customizations;
@@ -213,12 +245,28 @@ class ModuleLoader {
     }
   }
 
-  async eval(source, url, isEntryPoint = false) {
+  /**
+   *
+   * @param {string} source Source code of the module.
+   * @param {string} url URL of the module.
+   * @param {{ isMain?: boolean }|undefined} context - context object containing module metadata.
+   * @returns {object} The module wrap object.
+   */
+  createModuleWrap(source, url, context = kEmptyObject) {
+    return compileSourceTextModule(url, source, this, context);
+  }
+
+  /**
+   *
+   * @param {string} url URL of the module.
+   * @param {object} wrap Module wrap object.
+   * @param {boolean} isEntryPoint Whether the module is the entry point.
+   * @returns {Promise<object>} The module object.
+   */
+  async executeModuleJob(url, wrap, isEntryPoint = false) {
     const { ModuleJob } = require('internal/modules/esm/module_job');
-    const wrap = compileSourceTextModule(url, source, this);
     const module = await onImport.tracePromise(async () => {
-      const job = new ModuleJob(
-        this, url, undefined, wrap, false, false);
+      const job = new ModuleJob(this, url, undefined, wrap, kEvaluationPhase, false, false);
       this.loadCache.set(url, undefined, job);
       const { module } = await job.run(isEntryPoint);
       return module;
@@ -235,18 +283,32 @@ class ModuleLoader {
     };
   }
 
+  /**
+   *
+   * @param {string} source Source code of the module.
+   * @param {string} url URL of the module.
+   * @param {boolean} isEntryPoint Whether the module is the entry point.
+   * @returns {Promise<object>} The module object.
+   */
+  eval(source, url, isEntryPoint = false) {
+    const context = isEntryPoint ? { isMain: true } : undefined;
+    const wrap = this.createModuleWrap(source, url, context);
+    return this.executeModuleJob(url, wrap, isEntryPoint);
+  }
+
   /**
    * Get a (possibly not yet fully linked) module job from the cache, or create one and return its Promise.
    * @param {string} specifier The module request of the module to be resolved. Typically, what's
-   *                           requested by `import '<specifier>'` or `import('<specifier>')`.
+   *   requested by `import '<specifier>'` or `import('<specifier>')`.
    * @param {string} [parentURL] The URL of the module where the module request is initiated.
-   *                             It's undefined if it's from the root module.
+   *   It's undefined if it's from the root module.
    * @param {ImportAttributes} importAttributes Attributes from the import statement or expression.
-   * @returns {Promise<ModuleJobBase}
+   * @param {number} phase Import phase.
+   * @returns {Promise<ModuleJobBase>}
    */
-  async getModuleJobForImport(specifier, parentURL, importAttributes) {
+  async getModuleJobForImport(specifier, parentURL, importAttributes, phase) {
     const resolveResult = await this.resolve(specifier, parentURL, importAttributes);
-    return this.#getJobFromResolveResult(resolveResult, parentURL, importAttributes, false);
+    return this.#getJobFromResolveResult(resolveResult, parentURL, importAttributes, phase, false);
   }
 
   /**
@@ -256,11 +318,12 @@ class ModuleLoader {
    * @param {string} specifier See {@link getModuleJobForImport}
    * @param {string} [parentURL] See {@link getModuleJobForImport}
    * @param {ImportAttributes} importAttributes See {@link getModuleJobForImport}
-   * @returns {Promise<ModuleJobBase}
+   * @param {number} phase Import phase.
+   * @returns {Promise<ModuleJobBase>}
    */
-  getModuleJobForRequireInImportedCJS(specifier, parentURL, importAttributes) {
+  getModuleJobForRequireInImportedCJS(specifier, parentURL, importAttributes, phase) {
     const resolveResult = this.resolveSync(specifier, parentURL, importAttributes);
-    return this.#getJobFromResolveResult(resolveResult, parentURL, importAttributes, true);
+    return this.#getJobFromResolveResult(resolveResult, parentURL, importAttributes, phase, true);
   }
 
   /**
@@ -269,16 +332,21 @@ class ModuleLoader {
    * @param {{ format: string, url: string }} resolveResult Resolved module request.
    * @param {string} [parentURL] See {@link getModuleJobForImport}
    * @param {ImportAttributes} importAttributes See {@link getModuleJobForImport}
+   * @param {number} phase Import phase.
    * @param {boolean} isForRequireInImportedCJS Whether this is done for require() in imported CJS.
    * @returns {ModuleJobBase}
    */
-  #getJobFromResolveResult(resolveResult, parentURL, importAttributes, isForRequireInImportedCJS = false) {
+  #getJobFromResolveResult(resolveResult, parentURL, importAttributes, phase,
+                           isForRequireInImportedCJS = false) {
     const { url, format } = resolveResult;
     const resolvedImportAttributes = resolveResult.importAttributes ?? importAttributes;
     let job = this.loadCache.get(url, resolvedImportAttributes.type);
 
     if (job === undefined) {
-      job = this.#createModuleJob(url, resolvedImportAttributes, parentURL, format, isForRequireInImportedCJS);
+      job = this.#createModuleJob(url, resolvedImportAttributes, phase, parentURL, format,
+                                  isForRequireInImportedCJS);
+    } else {
+      job.ensurePhase(phase);
     }
 
     return job;
@@ -292,7 +360,7 @@ class ModuleLoader {
    * @param {string} source Source code. TODO(joyeecheung): pass the raw buffer.
    * @param {string} isMain Whether this module is a main module.
    * @param {CJSModule|undefined} parent Parent module, if any.
-   * @returns {{wrap: ModuleWrap, namespace: ModuleNamespaceObject}}
+   * @returns {{wrap: ModuleWrap, namespace: import('internal/modules/esm/utils').ModuleNamespaceObject}}
    */
   importSyncForRequire(mod, filename, source, isMain, parent) {
     const url = pathToFileURL(filename).href;
@@ -307,20 +375,58 @@ class ModuleLoader {
     //    TODO(joyeecheung): ensure that imported synchronous graphs are evaluated
     //    synchronously so that any previously imported synchronous graph is already
     //    evaluated at this point.
+    // TODO(joyeecheung): add something similar to CJS loader's requireStack to help
+    // debugging the the problematic links in the graph for import.
+    debug('importSyncForRequire', parent?.filename, '->', filename, job);
     if (job !== undefined) {
       mod[kRequiredModuleSymbol] = job.module;
+      const parentFilename = urlToFilename(parent?.filename);
+      // TODO(node:55782): this race may stop to happen when the ESM resolution and loading become synchronous.
+      if (!job.module) {
+        assert.fail(getRaceMessage(filename, parentFilename));
+      }
       if (job.module.async) {
-        throw new ERR_REQUIRE_ASYNC_MODULE();
+        throw new ERR_REQUIRE_ASYNC_MODULE(filename, parentFilename);
       }
-      if (job.module.getStatus() !== kEvaluated) {
-        const parentFilename = urlToFilename(parent?.filename);
-        let message = `Cannot require() ES Module ${filename} in a cycle.`;
-        if (parentFilename) {
-          message += ` (from ${parentFilename})`;
-        }
-        throw new ERR_REQUIRE_CYCLE_MODULE(message);
+      const status = job.module.getStatus();
+      debug('Module status', job, status);
+      if (status === kEvaluated) {
+        return { wrap: job.module, namespace: job.module.getNamespaceSync(filename, parentFilename) };
+      } else if (status === kInstantiated) {
+        // When it's an async job cached by another import request,
+        // which has finished linking but has not started its
+        // evaluation because the async run() task would be later
+        // in line. Then start the evaluation now with runSync(), which
+        // is guaranteed to finish by the time the other run() get to it,
+        // and the other task would just get the cached evaluation results,
+        // similar to what would happen when both are async.
+        mod[kRequiredModuleSymbol] = job.module;
+        const { namespace } = job.runSync(parent);
+        return { wrap: job.module, namespace: namespace || job.module.getNamespace() };
+      } else if (status === kErrored) {
+        // If the module was previously imported and errored, throw the error.
+        throw job.module.getError();
       }
-      return { wrap: job.module, namespace: job.module.getNamespaceSync() };
+      // When the cached async job have already encountered a linking
+      // error that gets wrapped into a rejection, but is still later
+      // in line to throw on it, just unwrap and throw the linking error
+      // from require().
+      if (job.instantiated) {
+        throwIfPromiseRejected(job.instantiated);
+      }
+      if (status !== kEvaluating) {
+        assert.fail(`Unexpected module status ${status}. ` +
+                    getRaceMessage(filename, parentFilename));
+      }
+      let message = `Cannot require() ES Module ${filename} in a cycle.`;
+      if (parentFilename) {
+        message += ` (from ${parentFilename})`;
+      }
+      message += ' A cycle involving require(esm) is not allowed to maintain ';
+      message += 'invariants mandated by the ECMAScript specification. ';
+      message += 'Try making at least part of the dependency in the graph lazily loaded.';
+      throw new ERR_REQUIRE_CYCLE_MODULE(message);
+
     }
     // TODO(joyeecheung): refactor this so that we pre-parse in C++ and hit the
     // cache here, or use a carrier object to carry the compiled module script
@@ -329,10 +435,10 @@ class ModuleLoader {
     const inspectBrk = (isMain && getOptionValue('--inspect-brk'));
 
     const { ModuleJobSync } = require('internal/modules/esm/module_job');
-    job = new ModuleJobSync(this, url, kEmptyObject, wrap, isMain, inspectBrk);
+    job = new ModuleJobSync(this, url, kEmptyObject, wrap, kEvaluationPhase, isMain, inspectBrk);
     this.loadCache.set(url, kImplicitTypeAttribute, job);
     mod[kRequiredModuleSymbol] = job.module;
-    return { wrap: job.module, namespace: job.runSync().namespace };
+    return { wrap: job.module, namespace: job.runSync(parent).namespace };
   }
 
   /**
@@ -341,9 +447,10 @@ class ModuleLoader {
    * @param {string} specifier Specifier of the the imported module.
    * @param {string} parentURL Where the import comes from.
    * @param {object} importAttributes import attributes from the import statement.
+   * @param {number} phase The import phase.
    * @returns {ModuleJobBase}
    */
-  getModuleJobForRequire(specifier, parentURL, importAttributes) {
+  getModuleJobForRequire(specifier, parentURL, importAttributes, phase) {
     const parsed = URLParse(specifier);
     if (parsed != null) {
       const protocol = parsed.protocol;
@@ -374,6 +481,7 @@ class ModuleLoader {
         }
         throw new ERR_REQUIRE_CYCLE_MODULE(message);
       }
+      job.ensurePhase(phase);
       // Otherwise the module could be imported before but the evaluation may be already
       // completed (e.g. the require call is lazy) so it's okay. We will return the
       // module now and check asynchronicity of the entire graph later, after the
@@ -415,7 +523,7 @@ class ModuleLoader {
 
     const inspectBrk = (isMain && getOptionValue('--inspect-brk'));
     const { ModuleJobSync } = require('internal/modules/esm/module_job');
-    job = new ModuleJobSync(this, url, importAttributes, wrap, isMain, inspectBrk);
+    job = new ModuleJobSync(this, url, importAttributes, wrap, phase, isMain, inspectBrk);
 
     this.loadCache.set(url, importAttributes.type, job);
     return job;
@@ -426,7 +534,7 @@ class ModuleLoader {
    * but the translator may return the ModuleWrap in a Promise.
    * @param {string} url URL of the module to be translated.
    * @param {string} format Format of the module to be translated. This is used to find
-   *                        matching translators.
+   *   matching translators.
    * @param {ModuleSource} source Source of the module to be translated.
    * @param {boolean} isMain Whether the module to be translated is the entry point.
    * @returns {ModuleWrap | Promise<ModuleWrap>}
@@ -495,13 +603,14 @@ class ModuleLoader {
    * by the time this returns. Otherwise it may still have pending module requests.
    * @param {string} url The URL that was resolved for this module.
    * @param {ImportAttributes} importAttributes See {@link getModuleJobForImport}
+   * @param {number} phase Import phase.
    * @param {string} [parentURL] See {@link getModuleJobForImport}
    * @param {string} [format] The format hint possibly returned by the `resolve` hook
    * @param {boolean} isForRequireInImportedCJS Whether this module job is created for require()
-   *                                            in imported CJS.
+   *   in imported CJS.
    * @returns {ModuleJobBase} The (possibly pending) module job
    */
-  #createModuleJob(url, importAttributes, parentURL, format, isForRequireInImportedCJS) {
+  #createModuleJob(url, importAttributes, phase, parentURL, format, isForRequireInImportedCJS) {
     const context = { format, importAttributes };
 
     const isMain = parentURL === undefined;
@@ -527,6 +636,7 @@ class ModuleLoader {
       url,
       importAttributes,
       moduleOrModulePromise,
+      phase,
       isMain,
       inspectBrk,
       isForRequireInImportedCJS,
@@ -543,12 +653,19 @@ class ModuleLoader {
    * @param {string} specifier The first parameter of an `import()` expression.
    * @param {string} parentURL Path of the parent importing the module.
    * @param {Record<string, string>} importAttributes Validations for the
-   *                                                  module import.
+   *   module import.
+   * @param {number} [phase] The phase of the import.
+   * @param {boolean} [isEntryPoint] Whether this is the realm-level entry point.
    * @returns {Promise<ModuleExports>}
    */
-  async import(specifier, parentURL, importAttributes, isEntryPoint = false) {
+  async import(specifier, parentURL, importAttributes, phase = kEvaluationPhase, isEntryPoint = false) {
     return onImport.tracePromise(async () => {
-      const moduleJob = await this.getModuleJobForImport(specifier, parentURL, importAttributes);
+      const moduleJob = await this.getModuleJobForImport(specifier, parentURL, importAttributes,
+                                                         phase);
+      if (phase === kSourcePhase) {
+        const module = await moduleJob.modulePromise;
+        return module.getModuleSourceObject();
+      }
       const { module } = await moduleJob.run(isEntryPoint);
       return module.getNamespace();
     }, {
@@ -560,6 +677,7 @@ class ModuleLoader {
 
   /**
    * @see {@link CustomizedModuleLoader.register}
+   * @returns {any}
    */
   register(specifier, parentURL, data, transferList, isInternal) {
     if (!this.#customizations) {
@@ -576,10 +694,9 @@ class ModuleLoader {
    * Resolve a module request to a URL identifying the location of the module. Handles customization hooks,
    * if any.
    * @param {string|URL} specifier The module request of the module to be resolved. Typically, what's
-   *                               requested by `import specifier`, `import(specifier)` or
-   *                               `import.meta.resolve(specifier)`.
+   *   requested by `import specifier`, `import(specifier)` or `import.meta.resolve(specifier)`.
    * @param {string} [parentURL] The URL of the module where the module request is initiated.
-   *                             It's undefined if it's from the root module.
+   *   It's undefined if it's from the root module.
    * @param {ImportAttributes} importAttributes Attributes from the import statement or expression.
    * @returns {Promise<{format: string, url: string}>}
    */
@@ -592,24 +709,30 @@ class ModuleLoader {
     if (this.#customizations) {  // Only has module.register hooks.
       return this.#customizations.resolve(specifier, parentURL, importAttributes);
     }
-    return this.#cachedDefaultResolve(specifier, parentURL, importAttributes);
+    return this.#cachedDefaultResolve(specifier, {
+      __proto__: null,
+      conditions: this.#defaultConditions,
+      parentURL,
+      importAttributes,
+    });
   }
 
   /**
    * Either return a cached resolution, or perform the default resolution which is synchronous, and
    * cache the result.
    * @param {string} specifier See {@link resolve}.
-   * @param {string} [parentURL] See {@link resolve}.
-   * @param {ImportAttributes} importAttributes See {@link resolve}.
+   * @param {{ parentURL?: string, importAttributes: ImportAttributes, conditions?: string[]}} context
    * @returns {{ format: string, url: string }}
    */
-  #cachedDefaultResolve(specifier, parentURL, importAttributes) {
+  #cachedDefaultResolve(specifier, context) {
+    const { parentURL, importAttributes } = context;
     const requestKey = this.#resolveCache.serializeKey(specifier, importAttributes);
     const cachedResult = this.#resolveCache.get(requestKey, parentURL);
     if (cachedResult != null) {
       return cachedResult;
     }
-    const result = this.defaultResolve(specifier, parentURL, importAttributes);
+    defaultResolve ??= require('internal/modules/esm/resolve').defaultResolve;
+    const result = defaultResolve(specifier, context);
     this.#resolveCache.set(requestKey, parentURL, result);
     return result;
   }
@@ -637,20 +760,21 @@ class ModuleLoader {
    * This is the default resolve step for module.registerHooks(), which incorporates asynchronous hooks
    * from module.register() which are run in a blocking fashion for it to be synchronous.
    * @param {string|URL} specifier See {@link resolveSync}.
-   * @param {{ parentURL?: string, importAttributes: ImportAttributes}} context See {@link resolveSync}.
+   * @param {{ parentURL?: string, importAttributes: ImportAttributes, conditions?: string[]}} context
+   *   See {@link resolveSync}.
    * @returns {{ format: string, url: string }}
    */
   #resolveAndMaybeBlockOnLoaderThread(specifier, context) {
     if (this.#customizations) {
       return this.#customizations.resolveSync(specifier, context.parentURL, context.importAttributes);
     }
-    return this.#cachedDefaultResolve(specifier, context.parentURL, context.importAttributes);
+    return this.#cachedDefaultResolve(specifier, context);
   }
 
   /**
    * Similar to {@link resolve}, but the results are always synchronously returned. If there are any
    * asynchronous resolve hooks from module.register(), it will block until the results are returned
-   * from the loader thread for this to be synchornous.
+   * from the loader thread for this to be synchronous.
    * This is here to support `import.meta.resolve()`, `require()` in imported CJS, and
    * `module.registerHooks()` hooks.
    *
@@ -667,25 +791,12 @@ class ModuleLoader {
       return resolveWithHooks(specifier, parentURL, importAttributes, this.#defaultConditions,
                               this.#resolveAndMaybeBlockOnLoaderThread.bind(this));
     }
-    return this.#resolveAndMaybeBlockOnLoaderThread(specifier, { parentURL, importAttributes });
-  }
-
-  /**
-   * Our `defaultResolve` is synchronous and can be used in both
-   * `resolve` and `resolveSync`. This function is here just to avoid
-   * repeating the same code block twice in those functions.
-   */
-  defaultResolve(originalSpecifier, parentURL, importAttributes) {
-    defaultResolve ??= require('internal/modules/esm/resolve').defaultResolve;
-
-    const context = {
+    return this.#resolveAndMaybeBlockOnLoaderThread(specifier, {
       __proto__: null,
       conditions: this.#defaultConditions,
-      importAttributes,
       parentURL,
-    };
-
-    return defaultResolve(originalSpecifier, context);
+      importAttributes,
+    });
   }
 
   /**
@@ -693,9 +804,9 @@ class ModuleLoader {
    * if any.
    * @param {string} url The URL of the module to be loaded.
    * @param {object} context Metadata about the module
-   * @returns {Promise<{ format: ModuleFormat, source: ModuleSource }>}
+   * @returns {Promise<{ format: ModuleFormat, source: ModuleSource }> | { format: ModuleFormat, source: ModuleSource }}
    */
-  async load(url, context) {
+  load(url, context) {
     if (loadHooks.length) {
       // Has module.registerHooks() hooks, use the synchronous variant that can handle both hooks.
       return this.#loadSync(url, context);
@@ -783,11 +894,11 @@ class CustomizedModuleLoader {
   /**
    * Register some loader specifier.
    * @param {string} originalSpecifier The specified URL path of the loader to
-   *                                   be registered.
+   *   be registered.
    * @param {string} parentURL The parent URL from where the loader will be
-   *                           registered if using it package name as specifier
+   *   registered if using it package name as specifier
    * @param {any} [data] Arbitrary data to be passed from the custom loader
-   * (user-land) to the worker.
+   *   (user-land) to the worker.
    * @param {any[]} [transferList] Objects in `data` that are changing ownership
    * @param {boolean} [isInternal] For internal loaders that should not be publicly exposed.
    * @returns {{ format: string, url: URL['href'] }}
@@ -799,10 +910,10 @@ class CustomizedModuleLoader {
   /**
    * Resolve the location of the module.
    * @param {string} originalSpecifier The specified URL path of the module to
-   *                                   be resolved.
+   *   be resolved.
    * @param {string} [parentURL] The URL path of the module's parent.
    * @param {ImportAttributes} importAttributes Attributes from the import
-   *                                            statement or expression.
+   *   statement or expression.
    * @returns {{ format: string, url: URL['href'] }}
    */
   resolve(originalSpecifier, parentURL, importAttributes) {
@@ -912,7 +1023,7 @@ function getOrInitializeCascadedLoader() {
  * Register a single loader programmatically.
  * @param {string|URL} specifier
  * @param {string|URL} [parentURL] Base to use when resolving `specifier`; optional if
- * `specifier` is absolute. Same as `options.parentUrl`, just inline
+ *   `specifier` is absolute. Same as `options.parentUrl`, just inline
  * @param {object} [options] Additional options to apply, described below.
  * @param {string|URL} [options.parentURL] Base to use when resolving `specifier`
  * @param {any} [options.data] Arbitrary data passed to the loader's `initialize` hook
